from truss.patch.signature import calc_truss_signature


def test_calc_truss_signature(custom_model_truss_dir):
    sign = calc_truss_signature(custom_model_truss_dir)
    assert len(sign.content_hashes_by_path) > 0
    assert "config.yaml" in sign.content_hashes_by_path
    with (custom_model_truss_dir / "config.yaml").open() as config_file:
        assert config_file.read() == sign.config


def test_calc_truss_signature_ignore_data_dir(custom_model_data_dir):
    sign = calc_truss_signature(custom_model_data_dir, ["data/*"])
    assert (
        len(sign.content_hashes_by_path) > 0
        and "data/foo.bar" not in sign.content_hashes_by_path
    )
    assert "config.yaml" in sign.content_hashes_by_path



import importlib
import os
import sys
import time
from contextlib import contextmanager
from pathlib import Path
from typing import Any
from unittest.mock import Mock, patch

import pytest
import yaml


@pytest.fixture
def app_path(truss_container_fs: Path, helpers: Any):
    truss_container_app_path = truss_container_fs / "app"
    model_file_content = """
class Model:
    def __init__(self):
        self.load_count = 0
    def load(self):
        self.load_count += 1
        if self.load_count <= 2:
            raise RuntimeError('Simulated error')
    def predict(self, request):
        return request
    """
    with helpers.file_content(
        truss_container_app_path / "model" / "model.py",
        model_file_content,
    ), helpers.sys_path(truss_container_app_path):
        yield truss_container_app_path


# TODO: Make this test work
@pytest.mark.skip(
    reason="Succeeds when tests in this file are run alone, but fails with the whole suit"
)
def test_model_wrapper_load_error_once(app_path):
    if "model_wrapper" in sys.modules:
        model_wrapper_module = sys.modules["model_wrapper"]
        importlib.reload(model_wrapper_module)
    else:
        model_wrapper_module = importlib.import_module("model_wrapper")
    model_wraper_class = getattr(model_wrapper_module, "ModelWrapper")
    config = yaml.safe_load((app_path / "config.yaml").read_text())
    model_wrapper = model_wraper_class(config)
    model_wrapper.load()
    # Allow load thread to execute
    time.sleep(1)
    output = model_wrapper.predict({})
    assert output == {}
    assert model_wrapper._model.load_count == 3


# TODO: Make this test work
@pytest.mark.skip(
    reason="Succeeds when tests in this file are run alone, but fails with the whole suit"
)
def test_model_wrapper_load_error_more_than_allowed(app_path, helpers):
    with helpers.env_var("NUM_LOAD_RETRIES_TRUSS", "0"):
        if "model_wrapper" in sys.modules:
            model_wrapper_module = sys.modules["model_wrapper"]
            importlib.reload(model_wrapper_module)
        else:
            model_wrapper_module = importlib.import_module("model_wrapper")
        model_wraper_class = getattr(model_wrapper_module, "ModelWrapper")
        config = yaml.safe_load((app_path / "config.yaml").read_text())
        model_wrapper = model_wraper_class(config)
        model_wrapper.load()
        # Allow load thread to execute
        time.sleep(1)
        assert model_wrapper.load_failed()


@pytest.mark.integration
async def test_model_wrapper_streaming_timeout(app_path):
    if "model_wrapper" in sys.modules:
        model_wrapper_module = sys.modules["model_wrapper"]
        importlib.reload(model_wrapper_module)
    else:
        model_wrapper_module = importlib.import_module("model_wrapper")
    model_wraper_class = getattr(model_wrapper_module, "ModelWrapper")

    # Create an instance of ModelWrapper with streaming_read_timeout set to 5 seconds
    config = yaml.safe_load((app_path / "config.yaml").read_text())
    config["runtime"]["streaming_read_timeout"] = 5
    model_wrapper = model_wraper_class(config)
    model_wrapper.load()
    assert model_wrapper._config.get("runtime").get("streaming_read_timeout") == 5


@pytest.mark.asyncio
async def test_trt_llm_truss_init_extension(trt_llm_truss_container_fs, helpers):
    app_path = trt_llm_truss_container_fs / "app"
    packages_path = trt_llm_truss_container_fs / "packages"
    with _clear_model_load_modules(), helpers.sys_paths(app_path, packages_path):
        model_wrapper_module = importlib.import_module("model_wrapper")
        model_wrapper_class = getattr(model_wrapper_module, "ModelWrapper")
        config = yaml.safe_load((app_path / "config.yaml").read_text())
        mock_extension = Mock()
        mock_extension.load = Mock()
        with patch.object(
            model_wrapper_module, "_init_extension", return_value=mock_extension
        ) as mock_init_extension:
            model_wrapper = model_wrapper_class(config)
            model_wrapper.load()
            called_with_specific_extension = any(
                call_args[0][0] == "trt_llm"
                for call_args in mock_init_extension.call_args_list
            )
            assert (
                called_with_specific_extension
            ), "Expected extension_name was not called"


@pytest.mark.asyncio
async def test_trt_llm_truss_predict(trt_llm_truss_container_fs, helpers):
    app_path = trt_llm_truss_container_fs / "app"
    packages_path = trt_llm_truss_container_fs / "packages"
    with _clear_model_load_modules(), helpers.sys_paths(
        app_path, packages_path
    ), _change_directory(app_path):
        model_wrapper_module = importlib.import_module("model_wrapper")
        model_wrapper_class = getattr(model_wrapper_module, "ModelWrapper")
        config = yaml.safe_load((app_path / "config.yaml").read_text())

        expected_predict_response = "test"
        mock_predict_called = False

        async def mock_predict(return_value):
            nonlocal mock_predict_called
            mock_predict_called = True
            return expected_predict_response

        mock_engine = Mock(predict=mock_predict)
        mock_extension = Mock()
        mock_extension.load = Mock()
        mock_extension.model_args = Mock(return_value={"engine": mock_engine})
        with patch.object(
            model_wrapper_module, "_init_extension", return_value=mock_extension
        ):
            model_wrapper = model_wrapper_class(config)
            model_wrapper.load()
            resp = await model_wrapper.predict({})
            mock_extension.load.assert_called()
            mock_extension.model_args.assert_called()
            assert mock_predict_called
            assert resp == expected_predict_response


@pytest.mark.asyncio
async def test_trt_llm_truss_missing_model_py(trt_llm_truss_container_fs, helpers):
    app_path = trt_llm_truss_container_fs / "app"
    (app_path / "model" / "model.py").unlink()

    packages_path = trt_llm_truss_container_fs / "packages"
    with _clear_model_load_modules(), helpers.sys_paths(
        app_path, packages_path
    ), _change_directory(app_path):
        model_wrapper_module = importlib.import_module("model_wrapper")
        model_wrapper_class = getattr(model_wrapper_module, "ModelWrapper")
        config = yaml.safe_load((app_path / "config.yaml").read_text())

        expected_predict_response = "test"
        mock_predict_called = False

        async def mock_predict(return_value):
            nonlocal mock_predict_called
            mock_predict_called = True
            return expected_predict_response

        mock_engine = Mock(predict=mock_predict)
        mock_extension = Mock()
        mock_extension.load = Mock()
        mock_extension.model_override = Mock(return_value=mock_engine)
        with patch.object(
            model_wrapper_module, "_init_extension", return_value=mock_extension
        ):
            model_wrapper = model_wrapper_class(config)
            model_wrapper.load()
            resp = await model_wrapper.predict({})
            mock_extension.load.assert_called()
            mock_extension.model_override.assert_called()
            assert mock_predict_called
            assert resp == expected_predict_response


@contextmanager
def _change_directory(new_directory: Path):
    original_directory = os.getcwd()
    os.chdir(str(new_directory))
    try:
        yield
    finally:
        os.chdir(original_directory)


@contextmanager
def _clear_model_load_modules():
    """Clear dangling references to model and model_wrapper modules

    We do this before to clear any debris from before, and after to clean up
    after self. This is meant for cases where we simulate running a truss model
    in process, where these modules are loaded dyamically.
    """
    # These are left over by TrussModuleLoader used by local prediction tests.
    # TODO(pankaj) Find a way for TrussModuleLoader to clean up after itself.
    _remove_model_load_sys_modules()
    yield
    _remove_model_load_sys_modules()


def _remove_model_load_sys_modules():
    if "model" in sys.modules:
        del sys.modules["model"]
    if "model.model" in sys.modules:
        del sys.modules["model.model"]
    if "model_wrapper" in sys.modules:
        del sys.modules["model_wrapper"]



import logging
import typing
from typing import IO, List, Optional, Tuple

import truss
from truss.remote.baseten import types as b10_types
from truss.remote.baseten.api import BasetenApi
from truss.remote.baseten.error import ApiError
from truss.remote.baseten.utils.tar import create_tar_with_progress_bar
from truss.remote.baseten.utils.transfer import multipart_upload_boto3
from truss.truss_handle import TrussHandle
from truss.util.path import load_trussignore_patterns

logger = logging.getLogger(__name__)


DEPLOYING_STATUSES = ["BUILDING", "DEPLOYING", "LOADING_MODEL", "UPDATING"]
ACTIVE_STATUS = "ACTIVE"


class ChainDeploymentHandle(typing.NamedTuple):
    chain_id: str
    chain_deployment_id: str


class ModelIdentifier:
    value: str


class ModelName(ModelIdentifier):
    def __init__(self, name: str):
        self.value = name


class ModelId(ModelIdentifier):
    def __init__(self, model_id: str):
        self.value = model_id


class ModelVersionId(ModelIdentifier):
    def __init__(self, model_version_id: str):
        self.value = model_version_id


def get_chain_id_by_name(api: BasetenApi, chain_name: str) -> Optional[str]:
    """
    Check if a chain with the given name exists in the Baseten remote.

    Args:
        api: BasetenApi instance
        chain_name: Name of the chain to check for existence

    Returns:
        chain_id if present, otherwise None
    """
    chains = api.get_chains()

    chain_name_id_mapping = {chain["name"]: chain["id"] for chain in chains}
    return chain_name_id_mapping.get(chain_name)


def exists_model(api: BasetenApi, model_name: str) -> Optional[str]:
    """
    Check if a model with the given name exists in the Baseten remote.

    Args:
        api: BasetenApi instance
        model_name: Name of the model to check for existence

    Returns:
        model_id if present, otherwise None
    """
    try:
        model = api.get_model(model_name)
    except ApiError as e:
        if (
            e.graphql_error_code
            == BasetenApi.GraphQLErrorCodes.RESOURCE_NOT_FOUND.value
        ):
            return None

        raise e

    return model["model"]["id"]


def create_chain(
    api: BasetenApi,
    chain_id: Optional[str],
    chain_name: str,
    chainlets: List[b10_types.ChainletData],
    is_draft: bool = False,
) -> ChainDeploymentHandle:
    if is_draft:
        response = api.deploy_draft_chain(chain_name, chainlets)
    elif chain_id:
        response = api.deploy_chain_deployment(chain_id, chainlets)
    else:
        response = api.deploy_chain(chain_name, chainlets)

    return ChainDeploymentHandle(
        chain_id=response["chain_id"],
        chain_deployment_id=response["chain_deployment_id"],
    )


def get_model_versions(api: BasetenApi, model_name: ModelName) -> Tuple[str, List]:
    query_result = api.get_model(model_name.value)["model"]
    return (query_result["id"], query_result["versions"])


def get_dev_version_from_versions(versions: List[dict]) -> Optional[dict]:
    """Given a list of model version dicts, returns the development version.

    Args:
        versions: List of model version dicts from the Baseten API

    Returns:
        The version in versions corresponding to the development model, or None
        if no development model exists
    """
    for version in versions:
        if version["is_draft"] is True:
            return version
    return None


def get_dev_version(api: BasetenApi, model_name: str) -> Optional[dict]:
    """Queries the Baseten API and returns a dict representing the
    development version for the given model_name.

    Args:
        api: BasetenApi instance
        model_name: Name of the model to get the development version for

    Returns:
        The development version of the model
    """
    model = api.get_model(model_name)
    versions = model["model"]["versions"]
    return get_dev_version_from_versions(versions)


def get_prod_version_from_versions(versions: List[dict]) -> Optional[dict]:
    # Loop over versions instead of using the primary_version field because
    # primary_version is set to the development version ID if no published
    # models exist.
    for version in versions:
        if version["is_primary"]:
            return version
    return None


def archive_truss(truss_handle: TrussHandle) -> IO:
    """
    Archive a TrussHandle into a tar file.

    Args:
        truss_handle: TrussHandle to archive

    Returns:
        A file-like object containing the tar file
    """
    truss_dir = truss_handle._spec.truss_dir
    ignore_patterns = []

    # check for a truss_ignore file and read the ignore patterns if it exists
    truss_ignore_file = truss_dir / ".truss_ignore"
    if truss_ignore_file.exists():
        ignore_patterns = load_trussignore_patterns(truss_ignore_file=truss_ignore_file)
    else:
        ignore_patterns = load_trussignore_patterns()

    try:
        temp_file = create_tar_with_progress_bar(truss_dir, ignore_patterns)
    except PermissionError:
        # workaround for Windows bug with Tempfile that causes PermissionErrors
        temp_file = create_tar_with_progress_bar(
            truss_dir, ignore_patterns, delete=False
        )
    temp_file.file.seek(0)
    return temp_file


def upload_truss(api: BasetenApi, serialize_file: IO) -> str:
    """
    Upload a TrussHandle to the Baseten remote.

    Args:
        api: BasetenApi instance
        serialize_file: File-like object containing the serialized TrussHandle

    Returns:
        The S3 key of the uploaded file
    """
    temp_credentials_s3_upload = api.model_s3_upload_credentials()
    s3_key = temp_credentials_s3_upload.pop("s3_key")
    s3_bucket = temp_credentials_s3_upload.pop("s3_bucket")
    multipart_upload_boto3(
        serialize_file.name, s3_bucket, s3_key, temp_credentials_s3_upload
    )
    return s3_key


def create_truss_service(
    api: BasetenApi,
    model_name: str,
    s3_key: str,
    config: str,
    semver_bump: str = "MINOR",
    is_trusted: bool = False,
    promote: bool = False,
    preserve_previous_prod_deployment: bool = False,
    is_draft: Optional[bool] = False,
    model_id: Optional[str] = None,
    deployment_name: Optional[str] = None,
    origin: Optional[b10_types.ModelOrigin] = None,
) -> Tuple[str, str]:
    """
    Create a model in the Baseten remote.

    Args:
        api: BasetenApi instance
        model_name: Name of the model to create
        s3_key: S3 key of the uploaded TrussHandle
        config: Base64 encoded JSON string of the Truss config
        semver_bump: Semver bump type, defaults to "MINOR"
        is_trusted: Whether the model is trusted, defaults to False
        promote: Whether to promote the model after deploy, defaults to False
        preserve_previous_prod_deployment: Wheter to scale old production deployment to zero
        deployment_name: Name to apply to the created deployment. Not applied to development model

    Returns:
        A tuple of the model ID and version ID
    """
    if is_draft:
        model_version_json = api.create_development_model_from_truss(
            model_name,
            s3_key,
            config,
            f"truss=={truss.version()}",
            is_trusted,
            origin=origin,
        )

        return (model_version_json["id"], model_version_json["version_id"])

    if model_id is None:
        model_version_json = api.create_model_from_truss(
            model_name=model_name,
            s3_key=s3_key,
            config=config,
            semver_bump=semver_bump,
            client_version=f"truss=={truss.version()}",
            is_trusted=is_trusted,
            deployment_name=deployment_name,
            origin=origin,
        )
        return (model_version_json["id"], model_version_json["version_id"])

    # Case where there is a model id already, create another version
    model_version_json = api.create_model_version_from_truss(
        model_id=model_id,
        s3_key=s3_key,
        config=config,
        semver_bump=semver_bump,
        client_version=f"truss=={truss.version()}",
        is_trusted=is_trusted,
        promote=promote,
        preserve_previous_prod_deployment=preserve_previous_prod_deployment,
        deployment_name=deployment_name,
    )
    model_version_id = model_version_json["id"]
    return (model_id, model_version_id)



from dataclasses import dataclass, field
from pathlib import Path
from typing import Dict

import yaml


@dataclass
class LocalConfig:
    secrets: Dict[str, str] = field(default_factory=dict)
    use_sudo: bool = False

    @staticmethod
    def from_dict(d):
        return LocalConfig(
            secrets=d.get("secrets", {}),
            use_sudo=d.get("use_sudo", False),
        )

    @staticmethod
    def from_yaml(yaml_path: Path):
        with yaml_path.open() as yaml_file:
            return LocalConfig.from_dict(yaml.safe_load(yaml_file))

    def to_dict(self):
        return {
            "secrets": self.secrets,
            "use_sudo": self.use_sudo,
        }

    def write_to_yaml_file(self, path: Path):
        with path.open("w") as config_file:
            yaml.dump(self.to_dict(), config_file)



# flake8: noqa
# type: ignore

# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""

import warnings

import briton_pb2 as briton__pb2
import grpc

GRPC_GENERATED_VERSION = "1.63.0"
GRPC_VERSION = grpc.__version__
EXPECTED_ERROR_RELEASE = "1.65.0"
SCHEDULED_RELEASE_DATE = "June 25, 2024"
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower

    _version_not_supported = first_version_is_lower(
        GRPC_VERSION, GRPC_GENERATED_VERSION
    )
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    warnings.warn(
        f"The grpc package installed is at version {GRPC_VERSION},"
        + f" but the generated code in briton_pb2_grpc.py depends on"
        + f" grpcio>={GRPC_GENERATED_VERSION}."
        + f" Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}"
        + f" or downgrade your generated code using grpcio-tools<={GRPC_VERSION}."
        + f" This warning will become an error in {EXPECTED_ERROR_RELEASE},"
        + f" scheduled for release on {SCHEDULED_RELEASE_DATE}.",
        RuntimeWarning,
    )


class BritonStub(object):
    """The greeting service definition."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Infer = channel.unary_stream(
            "/briton.Briton/Infer",
            request_serializer=briton__pb2.InferenceRequest.SerializeToString,
            response_deserializer=briton__pb2.InferenceAnswerPart.FromString,
            _registered_method=True,
        )


class BritonServicer(object):
    """The greeting service definition."""

    def Infer(self, request, context):
        """Sends a greeting"""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")


def add_BritonServicer_to_server(servicer, server):
    rpc_method_handlers = {
        "Infer": grpc.unary_stream_rpc_method_handler(
            servicer.Infer,
            request_deserializer=briton__pb2.InferenceRequest.FromString,
            response_serializer=briton__pb2.InferenceAnswerPart.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
        "briton.Briton", rpc_method_handlers
    )
    server.add_generic_rpc_handlers((generic_handler,))


# This class is part of an EXPERIMENTAL API.
class Briton(object):
    """The greeting service definition."""

    @staticmethod
    def Infer(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_stream(
            request,
            target,
            "/briton.Briton/Infer",
            briton__pb2.InferenceRequest.SerializeToString,
            briton__pb2.InferenceAnswerPart.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True,
        )
